跳到论文正文
W Papers独立论文阅读器 arXiv:2112.10223 论文正文 计算阅读时间 · 0%
引用这篇论文根据当前论文元数据生成,正式投稿前建议核对。

    
  
License: CC BY 4.0
arXiv:2112.10223v1 [cs.DC] 19 Dec 2021

Parallel Algorithms for
Adding a Collection of Sparse Matrices

Md Taufique Hussain1, Guttu Sai Abhishek2 Aydin Buluç3 and Ariful Azad4 Affiliation: 1 Indiana University, Bloomington, IN, USA (mth@indiana.edu) Affiliation: 2 IIT Bombay, India (180050036@iitb.ac.in) Affiliation: 3 Lawrence Berkeley National Laboratory, Berkeley, CA, USA (abuluc@lbl.gov) Affiliation: 4 Indiana University, Bloomington, IN, USA (azad@iu.edu)
Abstract

We develop a family of parallel algorithms for the SpKAdd operation that adds a collection of k sparse matrices. SpKAdd is a much needed operation in many applications including distributed memory sparse matrix-matrix multiplication (SpGEMM), streaming accumulations of graphs, and algorithmic sparsification of the gradient updates in deep learning. While adding two sparse matrices is a common operation in Matlab, Python, Intel MKL, and various GraphBLAS libraries, these implementations do not perform well when adding a large collection of sparse matrices. We develop a series of algorithms using tree merging, heap, sparse accumulator, hash table, and sliding hash table data structures. Among them, hash-based algorithms attain the theoretical lower bounds both on the computational and I/O complexities and perform the best in practice. The newly-developed hash SpKAdd makes the computation of a distributed-memory SpGEMM algorithm at least 2x faster than that the previous state-of-the-art algorithms.

I Introduction

Sparse matrices power many modern scientific codes, including those that arise in simulations, data analytics, and machine learning. An overlooked operation that is common across many applications is the summation (or reduction) of a collection of sparse matrices. In the deep learning community, this is called the sparse allreduce problem [1, 2] that arises due to algorithmic sparsification of the gradient updates for reduced communication. While most papers describe the problem as the reduction of sparse vectors, each holding gradient updates from one processor, in practice the computation becomes the reduction of a collection of sparse matrices due to mini-batching commonly used in gradient-based deep learning training. The existing work on this sparse allreduce problem is exclusively on the communication and distributed-memory aspects, leaving the actual in-node reduction itself understudied.

The use case that primarily motivated our work is the distributed-memory multiplication of two sparse matrices (SpGEMM). In virtually all algorithms for the distributed memory SpGEMM, there is a step where each processor has to sum intermediate products of the form ∑i=1k𝐀i​𝐁i\sum_{i=1}^{k}\mathbf{A}_{i}\mathbf{B}_{i}. We call this operation SpKAdd. While local SpGEMM algorithms have seen significant advances in their performance within the last 15 years [3], SpKAdd has not received any attention, leaving it a potential bottleneck. This is especially true when we try to achieve strong scaling or when there is overdecomposition of the problem for ease of load balancing: the more processes perform distributed SpGEMM, the more computation shifts from local SpGEMMs to SpKAdd. This seemingly counterintuitive fact is easy to verify when considering the extreme case where each process owns a single 1×11{\times}1 matrix and there are n2n^{2} processors for multiplying two n×nn{\times}n sparse matrices: SpKAdd does half the flops (i.e., all the additions) needed for the distributed SpGEMM. Furthermore, communication-avoiding SpGEMM algorithms [4, 5] utilize SpKAdd at two different phases: one within each 2D grid of the overall 3D process grid and another when reducing results across different 2D grids.

Finally, SpKAdd is needed for assembling local finite-element matrices into a global matrix [6]. This problem has traditionally been labeled as one that presents few opportunities for parallelism. In this work, we show this not to be true. In fact, SpKAdd has plenty of parallelism. However, existing implementations use suboptimal data structures for the accumulation, hurting their performance. In this paper, we systematically evaluate various different data structures and methods for SpKAdd in shared memory.

The main contribution of this paper is to develop efficient parallel SpKAdd algorithms that are work-efficient, use spatial locality when accessing memory, and do not require thread synchronization. We achieve these goals by utilizing heap, sparse accumulator (SPA) [7], hash table data structures. We designed a sliding hash algorithm after carefully considering (a) the data access patterns for different data structures, (b) matrix shapes and sparsity patterns, and (c) hardware features such as the cache size. We theoretically and experimentally demonstrate that hash and sliding hash based SpKAdd algorithms run the fastest for input matrices with various shapes and sparsity patterns on three different shared-memory platforms. We demonstrate the impact of fast SpKAdd algorithms by plugging the hash algorithm in a distributed SpGEMM implementation in the CombBLAS library [8]. When running SpGEMM on 2K nodes of a Cray XC40 supercomputer, the hash-based SpKAdd made the computation of the distributed SpGEMM 2×2\times faster than the previous implementation.

II Algorithmic landscape of the SpKAdd operation

II-A The problem

We consider the operation SpKAdd that adds kk sparse matrices 𝐀1,𝐀2,…,𝐀k\mathbf{A}_{1},\mathbf{A}_{2},...,\mathbf{A}_{k} and returns the summed (potentially sparse) matrix 𝐁=∑i=1k𝐀i\mathbf{B}{=}\sum_{i=1}^{k}\mathbf{A}_{i}, where 𝐀i∈ℝm×n\mathbf{A}_{i}{\in}\mathbb{R}^{m\times n} and 𝐁∈ℝm×n\mathbf{B}{\in}\mathbb{R}^{m\times n}. We assume that all input matrices 𝐀1,𝐀2,…,𝐀k\mathbf{A}_{1},\mathbf{A}_{2},...,\mathbf{A}_{k} are available in memory before performing the SpKAdd operation. Let nnz​(𝐀i)\textit{nnz}(\mathbf{A}_{i}) denote the number of nonzeros in 𝐀i\mathbf{A}_{i}. Then, nnz​(𝐁)≤∑i=1knnz​(𝐀i)\textit{nnz}(\mathbf{B})\leq\sum_{i=1}^{k}\textit{nnz}(\mathbf{A}_{i}). The compression factor 𝑐𝑓\mathop{\mathit{cf}}\nolimits of an SpKAdd operation is defined as ∑i=1knnz​(𝐀i)/nnz​(𝐁)\sum_{i=1}^{k}\textit{nnz}(\mathbf{A}_{i})/\textit{nnz}(\mathbf{B}), where 𝑐𝑓≥1\mathop{\mathit{cf}}\nolimits{\geq}1. When k=2k{=}2, the operation simply adds two sparse matrices. In this special case of k=2k{=}2, SpKAdd is equivalent to mkl_sparse_d_add in MKL, the “+” operator in Matlab and Python (with scipy sparse matrices as operands). While adding two sparse matrices sequentially or in parallel is a widely-implemented operation in any sparse matrix library, SpKAdd is rarely implemented in most libraries. One can repeatedly add sparse matrices in pairs to add kk sparse matrices. However, we will analytically and experimentally demonstrate that the pairwise addition could be significantly slower than a unified SpKAdd operation when kk is large.

Throughout the paper, we assume that input and output matrices are represented in the compressed sparse column (CSC) format that stores nonzero entries column by column. However, all algorithms discussed in this paper are equally applicable to compressed sparse row (CSR), coordinate (COO), doubly compressed, and other sparse matrix formats [9].

Algorithm 1 SpKAdd using 2-way addition of columns.

Input: A list of kk sparse matrices 𝐀1,…,𝐀k\mathbf{A}_{1},...,\mathbf{A}_{k}. Output: 𝐁\mathbf{B}. All matrices are in ℝm×n\mathbb{R}^{m\times n}


1: procedure SpKAdd2Way(𝐀1,…,𝐀k\mathbf{A}_{1},...,\mathbf{A}_{k})
2:   𝐁←𝐀1\mathbf{B}\leftarrow\mathbf{A}_{1}
3:   for i←2i\leftarrow 2 to kk do
4:    for j←1j\leftarrow 1 to nn do
5:      𝐁(:,j)←\mathbf{B}(:,j)\leftarrow ColAdd (𝐁(:,j),𝐀i(:,j)\mathbf{B}(:,j),\mathbf{A}_{i}(:,j))      
6:   return 𝐁\mathbf{B}

II-B SpKAdd using 2-way additions

At first, we describe SpKAdd algorithms that rely on 2-way addition of a pair of matrices. We have implemented parallel 2-way additions in our library, but an implementation from MKL, Matlab, and GraphBLAS [10, 11] can be used.

II-B1 SpKAdd using 2-way incremental additions

The simplest approach to implement SpKAdd is to incrementally add kk sparse matrices in pairs. Algorithm 1 provides a skeleton of SpKAdd implemented with 2-way incremental additions. Here, the iith iteration of the outer loop adds the current partial result with the next input matrix 𝐀i\mathbf{A}_{i}. Since matrices are stored in the CSC format, the jjth columns of 𝐀i\mathbf{A}_{i} and 𝐁\mathbf{B} can be added independently as denoted by the ColAdd operation in line 5 of Algorithm 1. Here, the jjth columns of a matrix is an array of (rowid, val) tuples where rowid denotes the row index of a nonzero entry. If columns are sorted in the ascending order of row index, the ColAdd function simply merges two lists of tuples, which is similar to the merging operation of the merge sort algorithm. Fig. 1(b) shows an example of 2-way incremental addition for the jjth columns of four matrices shown in Fig. 1(a).

Computational complexity. First, consider the cost of adding two matrices 𝐀1\mathbf{A}_{1} and 𝐀2\mathbf{A}_{2}. If each columns are sorted based on their row indices, a linear merging algorithm has O⁡(nnz​(𝐀1)+nnz​(𝐀2))O(\textit{nnz}(\mathbf{A}_{1})+\textit{nnz}(\mathbf{A}_{2})) complexity. Even though faster merging algorithm exists for imbalanced lists  [12], they often are too complicated for parallel implementations. In the worst case where the matrices have no common row indices, nnz​(𝐀1+𝐀2)=nnz​(𝐀1)+nnz​(𝐀2)\textit{nnz}(\mathbf{A}_{1}+\mathbf{A}_{2})=\textit{nnz}(\mathbf{A}_{1})+\textit{nnz}(\mathbf{A}_{2}). This is especially true when nonzeros in the input matrices are uniformly distributed such as in Erdős-Rényi (ER) random matrices. In this worst-case model, the iith iteration of the 2-way incremental algorithm has the cost of ∑l=1innz​(𝐀l)\sum_{l=1}^{i}{\textit{nnz}(\mathbf{A}_{l})} and the total cost of the algorithm is ∑i=2k∑l=1innz​(𝐀l)\sum_{i=2}^{k}\sum_{l=1}^{i}{\textit{nnz}(\mathbf{A}_{l})}. If all input are ER matrices with dd nonzeros per column, the overall computational cost is ∑i=2k∑l=1in​d=O⁡(k2​n​d)\sum_{i=2}^{k}\sum_{l=1}^{i}{nd}=O(k^{2}nd).

I/O complexity. In this paper, the I/O complexity measures the amount of data transferred to and from the main memory. If we assume that the output of every 2-way addition is stored in the main memory, the I/O complexity of the 2-way incremental algorithm matches the computational complexity discussed above. That means O⁡(k2​n​d)O(k^{2}nd) data is accessed from the main memory for ER matrices.

II-B2 SpKAdd using 2-way tree additions

The 2-way incremental addition is expensive both in terms of the computational complexity and data movements. One reason of this inefficiency is the imbalanced nature of the binary addition tree shown in Fig. 1(b) where the height of the tree is k−1k-1. To address this problem, we consider a balanced binary tree as shown in Fig. 1(c) where the height of the tree is lg⁡k\lg k. In the 2-way tree addition, input matrices are added in pairs in the leaves of the tree before the resultant matrices are added at internal nodes. Thus, the computational complexity of all additions at a given level of the binary tree is O⁡(∑i=1knnz​(𝐀i))O(\sum_{i=1}^{k}{\textit{nnz}(\mathbf{A}_{i})}), which results in the overall complexity of O⁡(lg⁡k​∑i=1knnz​(𝐀i))O(\lg k\sum_{i=1}^{k}{\textit{nnz}(\mathbf{A}_{i})}). Assuming that the intermediate matrices are stored in memory, the amount of data movement is also O⁡(lg⁡k​∑i=1knnz​(𝐀i))O(\lg k\sum_{i=1}^{k}{\textit{nnz}(\mathbf{A}_{i})}). For ER input matrices, the computational and data movement complexity become O⁡(n​d​k​lg⁡k)O(ndk\lg k). Hence, for any large value of kk, 2-way tree addition is faster than 2-way incremental addition both in theory and in practice. This performance improvement comes for free as we can still use any off-the-shelf function to add a pair of matrices.

Fig. 1: Data structure and algorithmic choices when adding the jjth columns of four input matrices. (a) Input and output columns, where each column stores (rowid, value) tuples. Different colors are used to show different input columns. (b) 2-way incremental addition. (c) 2-way tree addition where inputs are at the leaves of a binary tree. (d) k-way addition using a min-heap of size k=4k=4. The heap stores (rowid, matrixid, value) tuples and holds at most one entry from each input column. The entry with the minimum row index is at the root of the tree representing the min-heap. Only the first three heap operations (extracting the minimum entry and inserting another entry) are shown. We also show the current status of the output column. (e) k-way addition using a SPA of size m=8m=8. Any entry from the iith row of an input matrix such as 𝐀1​(i,j)\mathbf{A}_{1}(i,j) is added to the iith entry of the SPA. (f) k-way addition using a hash table of size 𝐁(:,j)=6\mathbf{B}(:,j)=6. For simplicity of depiction, we assume that the hash function maps row indices to hash table in the ascending order of row indices in 𝐁(:,j)\mathbf{B}(:,j) (e.g., 𝐁⁡(5,j)\mathbf{B}(5,j)) is the 4th entry of the output and mapped to the 4th position of the hash table).
Algorithm 2 SpKAdd using k-way addition of columns.

Input: A list of kk sparse matrices 𝐀1,…,𝐀k\mathbf{A}_{1},...,\mathbf{A}_{k}; Output: 𝐁\mathbf{B}. All matrices are in ℝm×n\mathbb{R}^{m\times n}


1: procedure SpKAdd(𝐀1,…,𝐀k\mathbf{A}_{1},...,\mathbf{A}_{k})
2:   for j←1j\leftarrow 1 to nn do
3:    𝐁(:,j)←\mathbf{B}(:,j)\leftarrow ColAddKWay (𝐀1(:,j),…,𝐀k(:,j)\mathbf{A}_{1}(:,j),...,\mathbf{A}_{k}(:,j)) ⊳\triangleright HeapAdd, SPAAdd, HashAdd, etc.   
4:   return 𝐁\mathbf{B}
Algorithm 3 SpKAdd using a min-heap. Only the addition of jjth columns is shown. The jjth column 𝐀(:,j)\mathbf{A}(:,j) is stored as a list of tuples (r,v)(r,v) where rr and vv denote row index and value.
1: procedure HeapAdd(𝐀1(:,j),…,𝐀k(:,j)\mathbf{A}_{1}(:,j),...,\mathbf{A}_{k}(:,j))
2:   𝐻𝑒𝑎𝑝←ϕ\mathop{\mathit{Heap}}\nolimits\leftarrow\phi
3:   for i←1i\leftarrow 1 to kk do
4:    (r,v)←(r,v)\leftarrow the nonzero entry with the smallest row index in 𝐀i(:,j)\mathbf{A}_{i}(:,j)
5:    Insert(𝐻𝑒𝑎𝑝\mathop{\mathit{Heap}}\nolimits, (r,i,v)(r,i,v))   
6:   while 𝐻𝑒𝑎𝑝\mathop{\mathit{Heap}}\nolimits ≠ϕ\neq\phi do
7:    (r,i,v)←(r,i,v)\leftarrow ExtractMin(𝐻𝑒𝑎𝑝\mathop{\mathit{Heap}}\nolimits)
8:    if 𝐁⁡(r,j)\mathbf{B}(r,j) exists then
9:      𝐁⁡(r,j)←𝐁⁡(r,j)+v\mathbf{B}(r,j)\leftarrow\mathbf{B}(r,j)+v
10:    else⊳\triangleright append at the end of the output
11:      𝐁(:,j)←𝐁(:,j)∪(r,v)\mathbf{B}(:,j)\leftarrow\mathbf{B}(:,j)\cup(r,v)    
12:    if 𝐀i\mathbf{A}_{i} still has unused entries then
13:      (rn​e​x​t,vn​e​x​t)←(r_{next},v_{next})\leftarrow the next element from 𝐀i\mathbf{A}_{i}
14:      Insert(𝐻𝑒𝑎𝑝\mathop{\mathit{Heap}}\nolimits, (rn​e​x​t,i,vn​e​x​t)(r_{next},i,v_{next}))      
15:   return 𝐁(:,j)\mathbf{B}(:,j)
Algorithm 4 SpKAdd using a SPA. Only the addition of jjth columns is shown. The jjth column 𝐀(:,j)\mathbf{A}(:,j) is stored as a list of tuples (r,v)(r,v) where rr and vv denote row index and value.
1: procedure SPAAdd(𝐀1(:,j),…,𝐀k(:,j)\mathbf{A}_{1}(:,j),...,\mathbf{A}_{k}(:,j))
2:   𝑆𝑃𝐴⁣←\mathop{\mathit{SPA}}\nolimits\leftarrow a dense array of length mm
3:   𝑖𝑑𝑥←ϕ\mathop{\mathit{idx}}\nolimits\leftarrow\phi ⊳\triangleright a list of valid indices in 𝑆𝑃𝐴\mathop{\mathit{SPA}}\nolimits
4:   for i←1i\leftarrow 1 to kk do
5:    for (r,v)∈𝐀i(:,j)(r,v)\in\mathbf{A}_{i}(:,j) do
6:      if r∈𝑖𝑑𝑥r\in\mathop{\mathit{idx}}\nolimits then 𝑆𝑃𝐴[r]←𝑆𝑃𝐴[r]+v\mathop{\mathit{SPA}}\nolimits[r]\leftarrow\mathop{\mathit{SPA}}\nolimits[r]+v
7:      else 𝑆𝑃𝐴[r]←v\mathop{\mathit{SPA}}\nolimits[r]\leftarrow v; 𝑖𝑑𝑥←𝑖𝑑𝑥∪r\mathop{\mathit{idx}}\nolimits\leftarrow\mathop{\mathit{idx}}\nolimits\cup r           
8:   Sort(𝑖𝑑𝑥\mathop{\mathit{idx}}\nolimits) ⊳\triangleright if sorted output is desired
9:   for r∈𝑖𝑑𝑥r\in\mathop{\mathit{idx}}\nolimits do
10:    append (r,𝑆𝑃𝐴[r])(r,\mathop{\mathit{SPA}}\nolimits[r]) into 𝐁(:,j)\mathbf{B}(:,j)   
11:   return 𝐁(:,j)\mathbf{B}(:,j)
Algorithm 5 SpKAdd using a hash table. Only the addition of jjth columns is shown. The jjth column 𝐀(:,j)\mathbf{A}(:,j) is stored as a list of tuples (r,v)(r,v) where rr and vv denote row index and value.
1: procedure HashAdd(𝐀1(:,j),…,𝐀k(:,j)\mathbf{A}_{1}(:,j),...,\mathbf{A}_{k}(:,j))
2:   𝐻𝑇⁣←\mathop{\mathit{HT}}\nolimits\leftarrow an array whose size is a power of two and greater than nnz(𝐁(:,j))\textit{nnz}(\mathbf{B}(:,j)); initialized with (−1,0)(-1,0)
3:   for i←1i\leftarrow 1 to kk do
4:    for (r,v)∈𝐀i(:,j)(r,v)\in\mathbf{A}_{i}(:,j) do
5:      h←Hash​(r)h\leftarrow\textsc{Hash}(r) ⊳\triangleright hash index from row index
6:      while true do
7:       if 𝐻𝑇⁡[h]\mathop{\mathit{HT[h]}}\nolimits is empty then
8:         𝐻𝑇[h]←(r,v)\mathop{\mathit{HT}}\nolimits[h]\leftarrow(r,v); break
9:       else if rr is found at 𝐻𝑇⁡[h]\mathop{\mathit{HT[h]}}\nolimits then
10:         add vv to the value at 𝐻𝑇[h]\mathop{\mathit{HT}}\nolimits[h]; break
11:       else⊳\triangleright hash conflict resolution
12:         h←h\leftarrow next hash index                 
13:   for every valid hash index h∈𝐻𝑇h\in\mathop{\mathit{HT}}\nolimits do
14:    append 𝐻𝑇[h]\mathop{\mathit{HT}}\nolimits[h] into 𝐁(:,j)\mathbf{B}(:,j)   
15:   Sort(𝐁(:,j)\mathbf{B}(:,j)) ⊳\triangleright if sorted output is desired
16:   return 𝐁(:,j)\mathbf{B}(:,j)

II-C SpKAdd using k-way additions

Since repeated use of 2-way additions leads to an inefficient SpKAdd operation, we developed several algorithms that use k-way additions. Algorithm 2 provides a template for such an algorithm where jjth columns of all input matrices are added to form the jjth column of the output. Line 3 of Algorithm 2 calls the k-way addition of columns for which we considered several data structures including heaps, sparse accumulators, and hash tables.

II-C1 HeapSpKAdd: k-way additions using a heap

Algorithm 3 shows the HeapAdd operation that uses a min-heap to add jjth columns of all input matrices. Here, 𝐻𝑒𝑎𝑝\mathop{\mathit{Heap}}\nolimits represents a binary tree storing kk tuples with each (r,i,v)(r,i,v) tuple represents v=𝐀i​(r,j)v=\mathbf{A}_{i}(r,j), i.e., vv is stored in the rrth row and jjth column of 𝐀i\mathbf{A}_{i}. Each tuple in the heap has a unique matrix index (the middle entry of the tuple), meaning that at most one tuple in the heap is from an input matrix. The min-heap uses row indices as the key and stores the tuple with the minimum row index at the root. Hence, extracting a tuple with the minimum row index takes O⁡(1)O(1) time, but inserting a new entry to the heap takes O⁡(lg⁡k)O(\lg k) time. See Fig. 1(d) for an example.

Lines 3-5 of Algorithm 3 initialize the heap with the smallest row indices from all input columns. Each iteration of the loop at line 6 extracts the minimum entry (r,i,v)(r,i,v) (based only on the row index) from the heap. If this is the first time we extract an entry from row rr, we append (r,v)(r,v) at the the end of the output. If there is already an entry in the output column at row rr, we simply add vv to this row index. Since the output column is formed in the ascending order of row indices, the existence of 𝐁⁡(r,j)\mathbf{B}(r,j) in line 8 can be checked in O⁡(1)O(1) time. If the minimum entry from the heap came from 𝐀i\mathbf{A}_{i}, we insert the next entry from 𝐀i\mathbf{A}_{i} into the heap (line 12-14). Here, the heap algorithm assumes that all columns of input matrices are sorted in the ascending order of row indices.

Complexity: Inserting a tuple to the heap takes O⁡(lg⁡k)O(\lg k) time. Since every nonzero entry from the input must be inserted to the heap for processing, the total computational complexity of the heap based algorithm is O⁡(lg⁡k​∑i=1knnz​(𝐀i))O(\lg k\sum_{i=1}^{k}{\textit{nnz}(\mathbf{A}_{i})}), which is equivalent to the 2-way tree algorithm algorithm. For I/O complexity, we assume that the heap (of size O⁡(k)O(k)) fits in a thread-private cache and HeapSpKAdd streams input columns from memory and writes the output column to memory without storing any intermediate matrices. Thus, the I/O cost of the heap algorithm is O⁡(∑i=1knnz​(𝐀i))O(\sum_{i=1}^{k}{\textit{nnz}(\mathbf{A}_{i})}). For ER matrices, the computational and I/O complexity of HeapSpKAdd become O⁡(n​d​k​lg⁡k)O(ndk\lg k) and O⁡(n​d​k)O(ndk), respectively.

II-C2 SPASpKAdd: k-way addition using a SPA

Algorithm 4 describes the addition of columns using a SPA. We represent SPA with two dense arrays 𝑆𝑃𝐴\mathop{\mathit{SPA}}\nolimits of length mm (number of rows in matrices) and 𝑖𝑑𝑥\mathop{\mathit{idx}}\nolimits of length nnz(𝐁(:,j))\textit{nnz}(\mathbf{B}(:,j)). The rrth entry of 𝑆𝑃𝐴\mathop{\mathit{SPA}}\nolimits stores the value at the rrth row of the output nnz​(𝐁​(r,j))\textit{nnz}(\mathbf{B}(r,j)). The 𝑖𝑑𝑥\mathop{\mathit{idx}}\nolimits array stores the indices of 𝑆𝑃𝐴\mathop{\mathit{SPA}}\nolimits with valid entries. Each iteration of the loop at line 4 in Algorithm 4 processes the input column from the ii matrix. Consider an entry (r,v)(r,v) from the current column 𝐀i(:,j)\mathbf{A}_{i}(:,j). If this row exists in the (\mathop{\mathit{(}}\nolimitsidx) array (i.e., a valid entry exists at 𝑆𝑃𝐴⁡[r]\mathop{\mathit{SPA[r]}}\nolimits), the value vv is added to the corresponding location at 𝑆𝑃𝐴\mathop{\mathit{SPA}}\nolimits (line 6). Otherwise, the row index rr is inserted into the 𝑖𝑑𝑥\mathop{\mathit{idx}}\nolimits array. If sorted output is desired the 𝑖𝑑𝑥\mathop{\mathit{idx}}\nolimits array is sorted before generating the result. See Fig. 1(e) for an example.

Complexity. Since every nonzero entry from the input is inserted once to SPA, the computational complexity of the SPA based algorithm is O⁡(∑i=1knnz​(𝐀i))O(\sum_{i=1}^{k}{\textit{nnz}(\mathbf{A}_{i})}). The I/O complexity is also O⁡(∑i=1knnz​(𝐀i))O(\sum_{i=1}^{k}{\textit{nnz}(\mathbf{A}_{i})}).

II-C3 HashSpKAdd: using k-way additions using a hash table

Algorithm 5 shows the HashAdd operation that uses a hash table 𝐻𝑇\mathop{\mathit{HT}}\nolimits to add jjth columns of all input matrices. The size of 𝐻𝑇\mathop{\mathit{HT}}\nolimits is a power of two and it is greater than nnz(B(:,j))\textit{nnz}(B(:,j)) (which is obtained via a symbolic phase discussed in Section II-D). The hash table 𝐻𝑇\mathop{\mathit{HT}}\nolimits stores r,vr,v tuples of row indices and values for all unique row indices in input columns. 𝐻𝑇\mathop{\mathit{HT}}\nolimits is initialized with (−1,0)(-1,0) where a negative row index denotes an unused entry in the hash table.

Each iteration of the loop at line 3 in Algorithm 5 processes the input column from the ii matrix. Consider an entry (r,v)(r,v) from the current column 𝐀i(:,j)\mathbf{A}_{i}(:,j). A suitable hash function is used to map the row index rr to a hash index hh (line 5). In our implementation, we used a multiplicative masking scheme Hash​(r)=(a∗r)&(2q−1)\textsc{Hash}(r)=(a*r)\&(2^{q}-1), where rr is the row index used as the hash key, aa is a prime number, 2q2^{q} is the size of the hash table with 2q≥nnz(B(:,j))2^{q}\geq\textit{nnz}(B(:,j)). The bitwise and operator (&\&) masks off the bottom qq bits so that the output of the hash function can be used as an index into a table of size 2q2^{q}. The while loop at line 6 search for the row index rr in 𝐻𝑇\mathop{\mathit{HT}}\nolimits. If such an entry is found, its value is updated in line 8. If not found (meaning that we encounter rr for the first time), the current tuple (r,v)(r,v) is stored in 𝐻𝑇\mathop{\mathit{HT}}\nolimits (line 7). When a valid entry with a different row index is encountered (line 11), it indicates a collision in hash tables. We used the linear probing scheme to resolve collisions. After we process all input columns, we insert all valid entries from the hash table to form the output. See Fig. 1(f) for an example.

Complexity. Inserting a value into a Hash table takes, on the average case, O⁡(1)O(1) time. Since every nonzero entry from the input is inserted to the hash table for processing, the average computational complexity of the hash based algorithm is O⁡(∑i=1knnz​(𝐀i))O(\sum_{i=1}^{k}{\textit{nnz}(\mathbf{A}_{i})}). The I/O cost of the hash algorithm is O⁡(∑i=1knnz​(𝐀i))O(\sum_{i=1}^{k}{\textit{nnz}(\mathbf{A}_{i})}). For ER matrices, both the computational and I/O complexity of HashSpKAdd become O⁡(n​d​k)O(ndk).

Algorithm 6 Computing nnz(B(:,j))\textit{nnz}(B(:,j)) using a hash table.
1: procedure HashSymbolic(𝐀1(:,j),…,𝐀k(:,j)\mathbf{A}_{1}(:,j),...,\mathbf{A}_{k}(:,j))
2:   𝐻𝑇⁣←\mathop{\mathit{HT}}\nolimits\leftarrow an array whose size is a power of two and greater than ∑i=1knnz​(Ai)\sum_{i=1}^{k}{\textit{nnz}(A_{i})}; initialized with −1-1
3:   nz ←\leftarrow 0
4:   for i←1i\leftarrow 1 to kk do
5:    for (r,v)∈𝐀i(:,j)(r,v)\in\mathbf{A}_{i}(:,j) do
6:      h←Hash​(r)h\leftarrow\textsc{Hash}(r) ⊳\triangleright hash index from row index
7:      while true do
8:       if 𝐻𝑇⁡[h]\mathop{\mathit{HT[h]}}\nolimits is empty then
9:         nz ←\leftarrow nz + 1
10:         𝐻𝑇[h]←r\mathop{\mathit{HT}}\nolimits[h]\leftarrow r; break
11:       else if rr is found at 𝐻𝑇⁡[h]\mathop{\mathit{HT[h]}}\nolimits then break
12:       else h←h\leftarrow next hash index                
13:   return nz

II-D The symbolic step to compute the output size

All k-way SpKAdd algorithms need to know nnz(𝐁(:,j))\textit{nnz}(\mathbf{B}(:,j)) for every column of the output to pre-allocate necessary memory for the output. This estimation is also needed to determine the hash table size. We pre-compute nnz(𝐁(:,j))\textit{nnz}(\mathbf{B}(:,j)) using a symbolic phase as shown in Algorithm 6. Here we use hash based symbolic phase, but heap and SPA could also be used to compute nnz(𝐁(:,j))\textit{nnz}(\mathbf{B}(:,j)). The hash table in the symbolic phase needs to store indices only. Algorithm 6 first creates a hash table of size ∑i=1knnz​(𝐀i)\sum_{i=1}^{k}\textit{nnz}(\mathbf{A}_{i}) and initializes it with -1. Then, similar to the actual addition (Algorithm 5), we search for the current row index rr in the hash table. The first time rr is found in the table, we increment the nonzero counter by one (lines 8-9). The computation and I/O complexity of Algorithm 6 is similar to that of Algorithm 5.

III Parallel and Cache Efficient SpKAdd Algorithms

III-A Parallel algorithms

The way we design 2-way and k-way SpKAdd algorithms makes it very easy to parallelize them. Since each column of 𝐁\mathbf{B} can be computed independently, we can parallelize line 4 of Algorithm 1 and line 2 of Algorithm 2. This parallelization strategy does not require any thread synchronization and it does not depend on the data structure (heap, hash, SPA, etc.) used to add the jjth columns of input matrices. In this case, we always execute Algorithms  3, 4 , 5, and 6 sequentially by a single thread. Note that parallelizing SpKAdd as a single multiway merging problem (e.g., when each matrix is stored as a list of (rowidx, colidx, val) tuples in the coordinate format) is not a trivial task as it requires the lists to be partitioned among threads [13]. Even though SpKAdd (when dividing columns among threads) is an embarrassingly parallel problem, the choices of data structures, matrix shapes and sparsity patterns, and the memory subsystem of the computing platform play significant roles on the attained performance of SpKAdd .

The impact of data structures on the parallel performance. Our parallel SpKAdd algorithms maintain thread-private data structures (because Algorithms  3, 4 , 5, and 6 are executed sequentially). Hence the total memory requirements for different data structures across TT threads are as follows: heap: O⁡(T⋅k)O(T\cdot k), SPA: O⁡(T⋅m)O(T\cdot m), Hash: O⁡(T⋅maxi=1n​nnz​(𝐁i))O(T\cdot\max_{i=1}^{n}{\textit{nnz}(\mathbf{B}_{i}})). The higher memory requirement of SPA may make the corresponding algorithm very slow because the algorithm accesses SPA at random locations. For example, if matrices have 32M rows, and 48 threads are used, the memory requirement for SPA could be more than 100GB. In contrast the memory requirement of heap is very small. However, the heap algorithm accesses all matrices concurrently (see Figure 1(d)). This could create bandwidth bottleneck if different matrices are stored in multiple sockets of a shared-memory platform.

Impact of matrix shapes and sparsity patterns on load imbalance. The computations in different columns of SpKAdd are fairly load balanced when nonzeros in input matrices are uniformly distributed such as in Erdős-Rényi matrices. However, for matrices with skewed nonzero distributions such as RMAT matrices, the computations can vary dramatically across columns. In the latter case, a static scheduling of threads hurts the parallel performance of the algorithm. In the symbolic phase we use total input non-zeros per column and in addition phase we use total output non-zeros per column to balance loads dynamically.

Algorithm 7 Computing nnz(B(:,j))\textit{nnz}(B(:,j)) using a sliding hash table. TT: number of threads; MM: total last-level cache; mm: number of rows in matrices; bb: number bytes needed for each entry in the hash table (4 bytes for the symbolic phase).
1: procedure SlHashSymbolic(𝐀1(:,j),…,𝐀k(:,j)\mathbf{A}_{1}(:,j),...,\mathbf{A}_{k}(:,j))
2:   𝑖𝑛𝑧←∑i=1knnz(Ai)\mathop{\mathit{inz}}\nolimits\leftarrow\sum_{i=1}^{k}{\textit{nnz}(A_{i})}
3:   𝑝𝑎𝑟𝑡𝑠←⌈(𝑖𝑛𝑧∗b∗T)/M⌉\mathop{\mathit{parts}}\nolimits\leftarrow\lceil(\mathop{\mathit{inz}}\nolimits*b*T)/M\rceil
4:   𝑛𝑧\mathop{\mathit{nz}}\nolimits ←0\leftarrow 0
5:   if 𝑝𝑎𝑟𝑡𝑠\mathop{\mathit{parts}}\nolimits=1 then ⊳\triangleright normal hash
6:    HashSymbolic(𝐀1(:,j),…,𝐀k(:,j)\mathbf{A}_{1}(:,j),...,\mathbf{A}_{k}(:,j))
7:   else⊳\triangleright sliding hash
8:    for i←1i\leftarrow 1 to 𝑝𝑎𝑟𝑡𝑠\mathop{\mathit{parts}}\nolimits do
9:      r1←i∗m/𝑝𝑎𝑟𝑡𝑠r_{1}\leftarrow i*m/\mathop{\mathit{parts}}\nolimits; r2←(i+1)∗m/𝑝𝑎𝑟𝑡𝑠r_{2}\leftarrow(i+1)*m/\mathop{\mathit{parts}}\nolimits
10:      𝑛𝑧\mathop{\mathit{nz}}\nolimits ←\leftarrow𝑛𝑧\mathop{\mathit{nz}}\nolimits + HashSymbolic(𝐀1(r1:r2,j),…,𝐀k(r1:r2,j)\mathbf{A}_{1}(r_{1}:r_{2},j),...,\mathbf{A}_{k}(r_{1}:r_{2},j))      
11:   return nz

III-B The Sliding Hash algorithm

In Algorithm 6 and Algorithm 5 we used a single hash table for the computation of the entire column of the output. These approaches work perfectly well as long as all hash tables used by all threads fit in the last level cache. To understand the memory requirements of hash tables, let MM be the size of last level cache in bytes, TT be the number of threads and bb be the number of bytes needed to store each entry in the hash table. When 32 bit indices and single-precision floating point values are stored, bb is 4 bytes for the symbolic phase and 8 bytes for the addition phase. Then, the total hash table memory required for the symbolic phase is MemSym = b∗T∗∑i=1knnz(𝐀i(:,j))b*T*\sum_{i=1}^{k}{\textit{nnz}(\mathbf{A}_{i}(:,j))} bytes. Similarly, the total hash table memory required for addition phase is MemAdd= b∗T∗nnz(𝐁(:,j))b*T*\textit{nnz}(\mathbf{B}(:,j)) bytes. If either MemSym or MemAdd is greater than MM, the hash table goes out of cache. Since hash tables are accessed randomly, out of cache accesses are quite expensive.

Algorithm 8 SpKAdd using a sliding hash table. TT: number of threads; MM: total last-level cache; mm: number of rows in matrices; bb: number bytes needed for each entry in the hash table (8 bytes for 32 bit indices and values).
1: procedure SlHashAdd(𝐀1(:,j),…,𝐀k(:,j)\mathbf{A}_{1}(:,j),...,\mathbf{A}_{k}(:,j))
2:   𝑜𝑛𝑧←SlHashSymbolic(𝐀1(:,j),…,𝐀k(:,j))\mathop{\mathit{onz}}\nolimits\leftarrow\textsc{SlHashSymbolic}(\mathbf{A}_{1}(:,j),...,\mathbf{A}_{k}(:,j))
3:   𝑝𝑎𝑟𝑡𝑠←⌈(𝑜𝑛𝑧∗b∗T)/M⌉\mathop{\mathit{parts}}\nolimits\leftarrow\lceil(\mathop{\mathit{onz}}\nolimits*b*T)/M\rceil
4:   𝑛𝑧\mathop{\mathit{nz}}\nolimits ←0\leftarrow 0
5:   if 𝑝𝑎𝑟𝑡𝑠\mathop{\mathit{parts}}\nolimits=1 then ⊳\triangleright normal hash
6:    HashAdd(𝐀1(:,j),…,𝐀k(:,j)\mathbf{A}_{1}(:,j),...,\mathbf{A}_{k}(:,j))
7:   else⊳\triangleright sliding hash
8:    for i←1i\leftarrow 1 to 𝑝𝑎𝑟𝑡𝑠\mathop{\mathit{parts}}\nolimits do
9:      r1←i∗m/𝑝𝑎𝑟𝑡𝑠r_{1}\leftarrow i*m/\mathop{\mathit{parts}}\nolimits ; r2←(i+1)∗m/𝑝𝑎𝑟𝑡𝑠r_{2}\leftarrow(i+1)*m/\mathop{\mathit{parts}}\nolimits
10:      𝐁(r1:r2,j)←\mathbf{B}(r_{1}{:}r_{2},j){\leftarrow} HashAdd(𝐀1(r1:r2,j),…,𝐀k(r1:r2,j)\mathbf{A}_{1}(r_{1}:r_{2},j),...,\mathbf{A}_{k}(r_{1}:r_{2},j))      
11:   return 𝐁(:,j)\mathbf{B}(:,j)
TABLE I: Summary of all SpKAdd algorithms when adding kk ER matrices each with dd nonzeros per column on average. TT denotes the number of threads in parallel algorithms and MM denotes the size of the last-level cache. The corresponding complexities for general matrices are shown in the main text.
Algorithm Work I/O (from memory) Data structure memory Data accesses Need sorted inputs?
2-way Incremental O⁡(k2​n​d)O(k^{2}nd) O⁡(k2​n​d)O(k^{2}nd) - streamed yes
2-way Tree O⁡(k​n​d​lg⁡k)O(knd\lg k) O⁡(k​n​d​lg⁡k)O(knd\lg k) - streamed yes
k-way Heap O⁡(k​n​d​lg⁡k)O(knd\lg k) O⁡(k​n​d)O(knd) O⁡(T​k)O(Tk) concurrent streaming of kk columns yes
k-way SPA O⁡(k​n​d)O(knd) O⁡(k​n​d)O(knd) O⁡(T​m)O(Tm) random access SPA no
k-way Hash O⁡(k​n​d)O(knd) O⁡(k​n​d)O(knd) O⁡(T​k​d)O(Tkd) random access hash table no
k-way Sliding Hash O⁡(k​n​d)O(knd) O⁡(k​n​d)O(knd) O⁡(M)O(M) in-cache hash table no

To address this problem, we designed a modified hash algorithm that limit the size of hash tables so that they fit within the last level cache. For example, when MemAdd is greater than MM, we create each hash table with no more than M/(b∗T)M/(b*T) entries so that all hash tables across threads fit in MM bytes. Since such a collection of hash tables cannot cover the entire output column, we slide a hash table along the column to produce outputs. Hence, we call it the sliding hash algorithm.

Algorithm 7 and Algorithm 8 describes the symbolic and SpKAdd algorithms using sliding hash tables. In Algorithm 7 we divide the required memory by MM to determine the number of parts (line 3). If the number of parts is one (i.e., all hash tables fit in MM byes), we simply call Algorithm 6 (lines 5-6). If more parts are needed, we partition rows equally (using binary searches) and then call Algorithm 6 for each part (lines 8-11). Similarly, Algorithm 7 identifies the memory requirement for all hash tables and uses the sliding hash strategy (lines 8-11) when needed. Thus, the sliding hash algorithms ensure that hash tables accesses are always performed within the cache, which gives significant performance boosts when the number of input matrices is large or when input matrices are denser. Note that hash tables in the symbolic phase could be larger than hash tables in the addition phase. For example, when c​fcf of an SpKAdd operation is 10, the symbolic phase may demand 10×10\times more memory for hash tables. Hence, the sliding hash table is more beneficial for the symbolic phase.

III-C Summary of all SpKAdd algorithms

Table I shows the complexity summary of algorithms when adding kk ER matrices each with dd nonzeros per column on average. We observe that 2-way and heap algorithms are not work efficient, and 2-way algorithms access more data from memory than other algorithms. Thus, we expect 2-way and heap algorithms to run slower than other algorithms. Our experimental results confirm this theoretical prediction. The memory requirements for various data structures also vary significantly. SPA requires the most memory, while the heap algorithm needs the least memory (assuming that k≪mk\ll m). The sliding hash algorithm allocates hash tables based on the available cache. Both SPA and hash algorithms accesses their data structures randomly. Random accesses can be a potential performance bottleneck if hash tables and SPA do not fit in the cache. The sliding hash table mitigates this problem by limiting the size of hash table. Finally, SPA and hash algorithms can operate with unsorted input columns, but 2-way and heap algorithms require sorted inputs.

TABLE II: Overview of the evaluation platforms.
Property
Intel
Skylake 8160
AMD
EPYC 7551
Cori Node
Intel KNL
Core Clock 2.10 GHz 2 GHz 1.4 GHz
L1 cache 32KB 32KB 32KB
L2 cache 1MB 512KB ×\times
LLC 32MB 8MB 34MB
Node Sockets 2 2 1
Cores/soc. 24 32 68
Memory 256GB 128GB 108GB
Fig. 2: Best performing algorithms with respect to the number of matrices (kk) and their sparsity (dd). Experiments were run on the Intel Skylake processor.

IV Results

IV-A Experimental setup

Experimental platforms. Most of our experiments are conducted on two different servers with Intel and AMD processors as described in Table II. To demonstrate the impact of our SpKAdd algorithms on a distributed SpGEMM algorithm, we ran some experiments on the KNL partition of Cori supercomputer at NERSC. We have implemented our in C/C++ programming language with OpenMP multi-threading. For distributed memory experiments, we plugged in our code in the sparse SUMMA algorithm available in the Combinatorial BLAS (CombBLAS) library [8].

Datasets. We use two types of matrices for the evaluation. We generate synthetic matrices using matrix generator, and also use publicly available protein similarity matrices. We use R-MAT [14], the recursive matrix generator, to generate two different non-zero patterns of synthetic matrices represented as ER and RMAT. ER matrix represents Erdős-Rényi  random graphs, and RMAT represents graphs with power-law degree distributions used for Graph500 benchmark. These matrices are generated with R-MAT seed parameters; a=b=c=d=0.25a{=}b{=}c{=}d{=}0.25 for ER matrix and a=0.57,b=c=0.19,d=0.05a{=}0.57,b{=}c{=}0.19,d{=}0.05 for G500 matrix. Most of our experiments use rectangular matrices with more rows than columns (m>nm>n). To generate kk RMAT matrices for SpKAdd experiments, we create a matrix m×nm\times n matrix and then split this matrix along the column to create kk m×m/km\times m/k matrices.

We also use protein-similarity networks Eukarya (3M rows, 3M columns, 360M nonzeros) and Isolates (35M rows, 35M columns, 17B nonzeros) that represent networks generated from isolate genomes in the IMG database and are publicly available with the HipMCL software [15]. We also used Metaclust50 (282M rows, 282M columns, 37B nonzeros) that stores similarities of proteins in Metaclust50 (https://metaclust.mmseqs.com/) dataset which contains predicted genes from metagenomes and metatranscriptomes of assembled contigs from IMG/Mand NCBI. We used these matrices to demonstrate the use of SpKAdd in distributed SpGEMM algorithms.

TABLE III: Runtime (sec) of different algorithms for different values of kk and different average nonzeros per column (dd) of ER matrices on the Intel Skylake processor (48 cores). Green cells represent the smallest runtime in each column.
Algorithm d=16d=16 d=1024d=1024 d=8192d=8192
k=4k=4 k=32k=32 k=128k=128 k=4k=4 k=32k=32 k=128k=128 k=4k=4 k=32k=32 k=128k=128
2-way Incremental 0.0022 0.0316 0.4506 0.0357 0.6618 5.7806 0.1746 2.7200 could not run
MKL Incremental 0.0588 0.6924 3.9971 0.1638 3.4144 29.1978 0.4717 12.9322 172.2600
2-way Tree 0.0019 0.0186 0.0832 0.0279 0.2440 1.2798 0.4416 0.8273 4.0545
MKL Tree 0.0499 0.5237 2.2155 0.1425 1.8711 8.2814 0.4121 5.4968 26.1705
Heap 0.0037 0.0112 0.0374 0.0730 0.5570 2.1732 0.5914 4.5293 16.1049
SPA 0.1237 0.1274 0.1309 0.1351 0.2844 0.8173 0.2579 1.3776 4.5269
Hash 0.0007 0.0016 0.0083 0.0110 0.1362 0.4463 0.1312 0.8062 4.3987
Sliding Hash 0.0021 0.0045 0.0162 0.0272 0.0936 0.3330 0.0993 0.5773 1.8096
TABLE IV: Runtime (sec) of different algorithms for different values of kk and different average nonzeros per column (dd) of RMAT matrices on the Intel Skylake processor (48 cores). Green cells represent the smallest runtime in each column.
Algorithm d=16d=16 d=64d=64 d=512d=512
k=4k=4 k=32k=32 k=128k=128 k=4k=4 k=32k=32 k=128k=128 k=4k=4 k=32k=32 k=128k=128
2-way Incremental 0.1155 0.8078 3.1727 0.3269 2.4913 9.3310 2.1307 could not run could not run
MKL Incremental 0.9839 6.0211 17.1921 3.5291 18.3133 47.9294 24.6052 109.5036 264.0963
2-way Tree 0.0774 0.2315 0.4691 0.2358 0.5673 0.9890 1.5213 3.3897 4.6937
MKL Tree 0.8670 2.2720 4.1899 2.9447 5.9594 8.3661 19.8781 31.0663 33.3237
Heap 0.1371 0.1384 0.1960 0.2422 1.0980 0.7603 1.8223 3.5277 6.7888
SPA 0.2467 0.2448 0.2386 0.5705 0.5659 0.6199 4.3047 4.6626 9.2565
Hash 0.1068 0.0739 0.0719 0.3240 0.3121 0.2925 1.7651 1.8187 1.8500
Sliding Hash 0.1251 0.0794 0.0762 0.3206 0.2625 0.2562 1.7909 1.7053 1.5471

IV-B Summary of performance with respect to sparsity patterns

At first, we study the best performing algorithms for different settings of input matrices including the number of matrices (kk), the average number of nonzeros (dd) in each column of a matrix, and their sparsity patterns (uniformly distribution in ER and scale-free distribution in RMAT).

The left subfigure in Fig. 2 shows that hash and sliding hash based SpKAdd algorithms are always the best performers for ER matrices. When the number of matrices is small or when the matrices are sparser, the hash tables from all threads fit within the cache. In these settings, there is no benefit of using the sliding hash algorithm. When the hash tables start to spill out from cache, the benefit of sliding hash tables becomes visible. The boundary between blue and orange regions is mostly determined by the last level cache size and the number threads sharing the cache. For example, when adding k=128k=128 matrices with average degree of d=512d=512, the expected number of nonzeros in an output column is 512*128=65,536. If each hash table entry needs 12 bytes to store (rowid, val) tuple and there are 48 threads, the total memory requirement to store all hash tables is about 38MB. Since our Intel Skylake processor has 32MB last level cache, the sliding hash is faster for k=128,d=512k=128,d=512. No other algorithms perform better than the hash SpKAdd for ER matrices.

The right subfigure in Fig. 2 shows that hash and sliding hash based SpKAdd algorithms are the best performers for RMAT matrices when k≥8k\geq 8. For k=4k=4, heap or 2-way tree algorithms perform the best. This is because the addition of RMAT matrices are dominated by denser columns. If we have just k=4k=4 columns to add and one of them is dense, 2-way addition algorithms can simply stream the denser column and add other sparse columns by utilising thread private L1 cache. With more matrices, the SpKAdd algorithms may need to add multiple denser columns in an iteration, which makes the hash algorithm perform better than its peers. Similar to the ER matrices, the performance of hash and sliding hash is determined by the size of the hash table. Sliding hash helps lower the runtime when the input matrices are denser and hash tables tend to go out of the cache.

Fig. 3: Strong scaling of various SpKAdd algorithms on the Intel processor.

While Fig. 2 provides an overall summary of the best performing algorithms, it does not show their actual runtimes. For a representative subset of matrices, we show the runtime of various algorithms in Table III and Table IV. These tables confirm the performance observed in Fig. 2. Additionally, we note few properties of various algorithms observed in Table III and Table IV. (a) The performance of the 2-way tree algorithm improves with matrix density. With a fixed k=128k=128 for ER matrices in Table III, the performance of 2-way tree relative to hash is: 10×10\times slower when d=16d=16, 3×3\times slower when d=1024d=1024, and 1.1×1.1\times faster when d=8196d=8196. Similarly, for RMAT matrices with a fixed k=128k=128 in Table IV, the performance of 2-way tree relative to hash is: 6×6\times slower when d=16d=16, 3×3\times slower when d=1024d=1024, and 2×2\times slower when d=8196d=8196. This is expected as hash tables become larger for denser matrices, making the 2-way tree algorithm more efficient. However, the sliding hash algorithm is still the fastest at the extreme setting with d=8196d=8196. (b) The SPA SpKAdd algorithm is as efficient as the hash SpKAdd algorithm for denser matrices as seen in the last column of Table III. This is expected because the hash table size (nnz(B(:,j))\textit{nnz}(B(:,j))) becomes O⁡(m)O(m) for dense output columns. This also suggests that the benefits of sliding hash can also be observed in the SPA algorithm if we partition the SPA array based on row indices [16]. (c) 2-way SpKAdd using MKL is significantly slower than other algorithms. The MKL-based 2-way SpKAdd algorithms do not show promising results in our experiments. This especially true for larger values of kk.

IV-C Scalability

We conducted the strong scaling experiments on the Intel Skylake processors with up to 48 threads (at most one thread per core). Fig. 3 shows the scalability results with (a) ER matrices, (b) RMAT matrices, and (c) 64 lower rank matrices generated when running a distributed SpGEMM algorithm with the Eukarya matrix. We observe that most algorithms scale almost linearly except 2-way SpKAdd algorithms (MKL Tree and 2-way tree) and SPA-based SpKAdd . Our load balancing technique ensures that k-way SpKAdd implementations scale well for matrices with skewed nonzero distributions. The performance of SPA suffers at high thread counts because of O⁡(T​m)O(Tm) memory requirement where TT is the number of threads and mm is the number of rows in matrices. However, the sequential SPA-SpKAdd performs better than other algorithms on a single thread because a SPA with 4M entries can still fit in the cache that is not shared among threads. However, for matrices with more rows (e.g., more that 32M rows), SPA would go out of cache even with a single thread.

Fig. 4: Determining optimum hash table size for the sliding hash algorithm. (a), (b), (c) and (d) are on the Intel Skylake processor. (e) and (f) are on AMD EPYC. The number of partitions (line 3 in Algorithm 8 and  7) depends on the selected hash table sizes. The rightmost points in all figures denote hash table sizes that need no partitioning of input columns.

IV-D Understanding the performance of the hash algorithms

Since hash and sliding hash algorithms are among the best performers, we aim to understand when and why they perform better than their peers. Fig. 4 shows the runtime of the sliding hash algorithm for different sizes of hash tables. The number of partitions (line 3 in Algorithm 8 and  7) depends on the selected hash table sizes. For example, in Fig. 4(b), the expected number of nonzeros across kk input columns is about 1M. Thus, a hash table with 16K entries will translates into 64 partitions, but a hash table with with 1M entries translates into one partition where no sliding hash is used.

TABLE V: Last level cache misses on Intel Skylake.
Cases (Fig. 4) LL miss read
Sliding Hash Hash
(a) 1.8M 1.4M
(b) 214M 734M
(c) 344M 409M
(d) 150M 152M

The optimum hash table sizes observed in Fig. 4 are related to the cache sizes in different processors. Fig. 4(a) is a case with very small matrices where the L1 cache plays a major role. Since our Intel processor has 32KB L1 cache per core, the L1 cache can hold a hash table with 4K entries (assuming that 8 bytes are needed for each entry in the hash table). That is why the optimum hash table size is observed near 4K in Fig. 4(a). For bigger matrices, the last level L3 cache influence the optimum hash table size. 32MB L3 cache shared among threads can hold hash tables with 64K entries. Hence, we observe the optimum hash table sizes near 64K in Fig. 4(b) and (c). The AMD processor used in our experiment has a smaller L3 cache (8MB). As a result, the optimum hash table size is smaller Fig. 4(e) and (f) that their corresponding sizes in Fig. 4(b) and (c). We experimentally validate the aforementioned relations between hash table and cache sizes by using the Cachegrind profiler. Table V reports the number of LL cache misses for the hash and sliding hash algorithms with various input matrices shown in Fig. 4. As expected, for cases (b) and (c), the sliding hash encounter fewer LL cache misses than the hash algorithm. However, the sliding hash shows no observable benefits for cases (a) and (d) because these cases do not require large hash tables. The compression factor SpKAdd also plays a role in the relative runtimes of the symbolic and addition phases of spkadd. For example, the Eukarya addition in Fig. 4(d) has a 𝑐𝑓=22.6\mathop{\mathit{cf}}\nolimits=22.6. Hence, the symbolic phase needed hash tables that are 27×27\times larger than the hash tables needed in the addition phase. Hence, the symbolic phase is very sensitive to the hash table size in Fig. 4(d), but the addition phase remains insensitive.

Fig. 5: The use of SpKAdd in distributed sparse SUMMA algorithm (with stationary output). Here, the first input matrix is distributed on a 2×42\times 4 process grid and the second matrix is stored on a 4×24\times 2 grid. 𝐀⁡(P0,0)\mathbf{A}(P_{0,0}) denotes the submatrix of 𝐀\mathbf{A} stored in process P0,0P_{0,0}. Local submatrices in different processes are shown with thick grids. For this configuration, SUMMA operates in four phases where input submatrices are broadcasted along the process grid row and columns. We show the computation at P0,0P_{0,0} where four local multiplications are performed, and the intermediate matrices are added using an SpKAdd operation.
Fig. 6: Effect of SpKAdd in distributed memory SpGEMM running on Cori KNL nodes. In both cases, square process grid and 8 threads per process were used.

IV-E Impact on the distributed SpGEMM algorithm

Among many applications of SpKAdd , we discuss an application in distributed memory SpGEMM algorithms based on sparse SUMMA [17]. Fig. 5 shows an example of sparse SUMMA with 8 processes. In this example, the first and second input matrices are distributed on a 2×42\times 4 and 4×24\times 2 logical process grids, respectively. After different stages of the algorithm, intermediate results from all stages are added to get the final results at different processes. The last computational step of this sparse SUMMA algorithm needs an SpKAdd operation as shown in Fig. 5 for process P0,0P_{0,0}.

To test the utility of our SpKAdd algorithms, we plugged our hash implementation in the SpGEMM code inside the CombBLAS library [8] and compare its performance with an existing heap SpKAdd implementation. Since sparse SUMMA has two computational kernels (local SpGEMM and SpKAdd ), we show the runtime of both computational steps by excluding the communication costs. Fig. 6 shows the result for two large scale distributed executions with 16384 processes for Metaclust50 and 4096 processes for Isolates-small respectively. Fig. 6 shows that the hash algorithms reduces the cost of SpKAdd by an order of magnitude when compared with the existing heap algorithm. Since hash SpKAdd does not need sorted inputs, the preceding local multiplication can avoid sorting intermediate results. Skipping sorting in the local multiplications can make it 20% faster as shown in Fig. 6.

V Conclusions

Despite being widely used in graph accumulation, matrix summarizing, and distributed-memory SpGEMM algorithms, SpKAdd algorithms are not well studied in the literature. In this paper, we theoretically and experimentally showed that SpKAdd implemented using off-the-shelf 2-way additions is not work efficient and accesses more data from memory. We develop a family of shared-memory parallel SpKAdd algorithms that rely on heap, SPA, and hash table data structures to add matrices column by column. Among these algorithms, a variant of hash algorithm is shown to be work efficient, uses spatial locality when accessing hash tables, and works with both sorted and unsorted input matrices. A limitation of the presented algorithms is our assumption that all input matrices fit in the main memory. When this assumption is not true (because the memory is limitated or matrices arrive in batches [18]), we can still arrange input matrices in multiple batches and then use SpKAdd for each batch. We leave streaming SpKAdd algorithms as a future work.

References

  • [1] H. Zhao and J. Canny, “Kylix: A sparse allreduce for commodity clusters,” in ICPP, 2014, pp. 273–282.
  • [2] C. Renggli, S. Ashkboos, M. Aghagolzadeh, D. Alistarh, and T. Hoefler, “SparCML: High-performance sparse communication for machine learning,” in SC, 2019, pp. 1–15.
  • [3] Y. Nagasaka, S. Matsuoka, A. Azad, and A. Buluc, “High-performance sparse matrix-matrix products on intel KNL and multicore architectures,” in ICPP Workshops. New York, NY, USA: ACM, 2018, pp. 34:1–34:10.
  • [4] A. Lazzaro, J. VandeVondele, J. Hutter, and O. Schütt, “Increasing the efficiency of sparse matrix-matrix multiplication with a 2.5 d algorithm and one-sided MPI,” in PASC, 2017, pp. 1–9.
  • [5] M. T. Hussain, O. Selvitopi, A. Buluç, and A. Azad, “Communication-avoiding and memory-constrained sparse matrix-matrix multiplication at extreme scale,” in IPDPS. IEEE, 2021.
  • [6] Z. Fu, T. J. Lewis, R. M. Kirby, and R. T. Whitaker, “Architecting the finite element method pipeline for the GPU,” Journal of computational and applied mathematics, vol. 257, pp. 195–211, 2014.
  • [7] J. R. Gilbert, C. Moler, and R. Schreiber, “Sparse matrices in matlab: Design and implementation,” SIAM journal on matrix analysis and applications, vol. 13, no. 1, pp. 333–356, 1992.
  • [8] A. Azad, O. Selvitopi, M. T. Hussain, J. Gilbert, and A. Buluc, “Combinatorial blas 2.0: Scaling combinatorial algorithms on distributed-memory systems,” IEEE TPDS, 2021.
  • [9] D. Langr and P. Tvrdik, “Evaluation criteria for sparse matrix storage formats,” IEEE TPDS, vol. 27, no. 2, pp. 428–440, 2015.
  • [10] A. Buluç, T. Mattson, S. McMillan, J. Moreira, and C. Yang, “Design of the GraphBLAS API for C,” in Proceedings of the IPDPSW. IEEE, 2017, pp. 643–652.
  • [11] T. A. Davis, “Algorithm 1000: SuiteSparse:GraphBLAS: Graph algorithms in the language of sparse linear algebra,” ACM Transactions on Mathematical Software (TOMS), vol. 45, no. 4, pp. 1–25, 2019.
  • [12] M. R. Brown and R. E. Tarjan, “A fast merging algorithm,” Journal of the ACM (JACM), vol. 26, no. 2, pp. 211–226, 1979.
  • [13] A. Azad, G. Ballard, A. Buluç, J. Demmel, L. Grigori, O. Schwartz, S. Toledo, and S. Williams, “Exploiting multiple levels of parallelism in sparse matrix-matrix multiplication,” SIAM Journal on Scientific Computing, vol. 38, no. 6, pp. C624–C651, 2016.
  • [14] D. Chakrabarti, Y. Zhan, and C. Faloutsos, “R-MAT: A recursive model for graph mining,” in SIAM SDM. SIAM, 2004, pp. 442–446.
  • [15] A. Azad, A. Buluç, G. A. Pavlopoulos, N. C. Kyrpides, and C. A. Ouzounis, “HipMCL: a high-performance parallel implementation of the Markov clustering algorithm for large-scale networks,” Nucleic Acids Research, vol. 46, no. 6, pp. e33–e33, 01 2018.
  • [16] M. M. A. Patwary, N. R. Satish, N. Sundaram, J. Park, M. J. Anderson, S. G. Vadlamudi, D. Das, S. G. Pudov, V. O. Pirogov, and P. Dubey, “Parallel efficient sparse matrix-matrix multiplication on multicore platforms,” in International Conference on High Performance Computing. Springer, 2015, pp. 48–57.
  • [17] A. Buluç and J. R. Gilbert, “Parallel sparse matrix-matrix multiplication and indexing: Implementation and experiments,” SIAM Journal on Scientific Computing, vol. 34, no. 4, pp. C170–C191, 2012.
  • [18] O. Selvitopi, M. T. Hussain, A. Azad, and A. Buluç, “Optimizing high performance Markov clustering for pre-exascale architectures,” in IPDPS, 2020.